@@ -11,7 +11,6 @@ from asyncio import (
     Future,
     Task,
     ensure_future,
-    get_event_loop,
     new_event_loop,
     set_event_loop,
     sleep,
@@ -42,13 +41,14 @@ from typing import (
 from prompt_toolkit.buffer import Buffer
 from prompt_toolkit.cache import SimpleCache
 from prompt_toolkit.clipboard import Clipboard, InMemoryClipboard
+from prompt_toolkit.cursor_shapes import AnyCursorShapeConfig, to_cursor_shape_config
 from prompt_toolkit.data_structures import Size
 from prompt_toolkit.enums import EditingMode
 from prompt_toolkit.eventloop import (
     get_traceback_from_context,
     run_in_executor_with_context,
 )
-from prompt_toolkit.eventloop.utils import call_soon_threadsafe
+from prompt_toolkit.eventloop.utils import call_soon_threadsafe, get_event_loop
 from prompt_toolkit.filters import Condition, Filter, FilterOrBool, to_filter
 from prompt_toolkit.formatted_text import AnyFormattedText
 from prompt_toolkit.input.base import Input
@@ -89,6 +89,7 @@ from prompt_toolkit.styles import (
 )
 from prompt_toolkit.utils import Event, in_main_thread
 
+from ..utils import is_windows
 from .current import get_app_session, set_app
 from .run_in_terminal import in_terminal, run_in_terminal
 
@@ -216,10 +217,11 @@ class Application(Generic[_AppResult]):
         max_render_postpone_time: Union[float, int, None] = 0.01,
         refresh_interval: Optional[float] = None,
         terminal_size_polling_interval: Optional[float] = 0.5,
-        on_reset: Optional[ApplicationEventHandler] = None,
-        on_invalidate: Optional[ApplicationEventHandler] = None,
-        before_render: Optional[ApplicationEventHandler] = None,
-        after_render: Optional[ApplicationEventHandler] = None,
+        cursor: AnyCursorShapeConfig = None,
+        on_reset: Optional["ApplicationEventHandler[_AppResult]"] = None,
+        on_invalidate: Optional["ApplicationEventHandler[_AppResult]"] = None,
+        before_render: Optional["ApplicationEventHandler[_AppResult]"] = None,
+        after_render: Optional["ApplicationEventHandler[_AppResult]"] = None,
         # I/O.
         input: Optional[Input] = None,
         output: Optional[Output] = None,
@@ -266,6 +268,8 @@ class Application(Generic[_AppResult]):
         self.refresh_interval = refresh_interval
         self.terminal_size_polling_interval = terminal_size_polling_interval
 
+        self.cursor = to_cursor_shape_config(cursor)
+
         # Events.
         self.on_invalidate = Event(self, on_invalidate)
         self.on_reset = Event(self, on_reset)
@@ -631,6 +635,8 @@ class Application(Generic[_AppResult]):
         self,
         pre_run: Optional[Callable[[], None]] = None,
         set_exception_handler: bool = True,
+        handle_sigint: bool = True,
+        slow_callback_duration: float = 0.5,
     ) -> _AppResult:
         """
         Run the prompt_toolkit :class:`~prompt_toolkit.application.Application`
@@ -647,9 +653,24 @@ class Application(Generic[_AppResult]):
         :param set_exception_handler: When set, in case of an exception, go out
             of the alternate screen and hide the application, display the
             exception, and wait for the user to press ENTER.
+        :param handle_sigint: Handle SIGINT signal if possible. This will call
+            the `<sigint>` key binding when a SIGINT is received. (This only
+            works in the main thread.)
+        :param slow_callback_duration: Display warnings if code scheduled in
+            the asyncio event loop takes more time than this. The asyncio
+            default of `0.1` is sometimes not sufficient on a slow system,
+            because exceptionally, the drawing of the app, which happens in the
+            event loop, can take a bit longer from time to time.
         """
         assert not self._is_running, "Application is already running."
 
+        if not in_main_thread() or is_windows():
+            # Handling signals in other threads is not supported.
+            # Also on Windows, `add_signal_handler(signal.SIGINT, ...)` raises
+            # `NotImplementedError`.
+            # See: https://github.com/prompt-toolkit/python-prompt-toolkit/issues/1553
+            handle_sigint = False
+
         async def _run_async() -> _AppResult:
             "Coroutine."
             loop = get_event_loop()
@@ -721,13 +742,14 @@ class Application(Generic[_AppResult]):
             with self.input.raw_mode(), self.input.attach(
                 read_from_input
             ), attach_winch_signal_handler(self._on_resize):
-                self.create_background_task(self._poll_output_size())
 
                 # Draw UI.
                 self._request_absolute_cursor_position()
                 self._redraw()
                 self._start_auto_refresh_task()
 
+                self.create_background_task(self._poll_output_size())
+
                 # Wait for UI to finish.
                 try:
                     result = await f
@@ -769,49 +791,68 @@ class Application(Generic[_AppResult]):
                         # Store unprocessed input as typeahead for next time.
                         store_typeahead(self.input, self.key_processor.empty_queue())
 
-                return result
+                return cast(_AppResult, result)
 
         async def _run_async2() -> _AppResult:
             self._is_running = True
-
-            # Make sure to set `_invalidated` to `False` to begin with,
-            # otherwise we're not going to paint anything. This can happen if
-            # this application had run before on a different event loop, and a
-            # paint was scheduled using `call_soon_threadsafe` with
-            # `max_postpone_time`.
-            self._invalidated = False
-
-            loop = get_event_loop()
-            if set_exception_handler:
-                previous_exc_handler = loop.get_exception_handler()
-                loop.set_exception_handler(self._handle_exception)
-
             try:
-                with set_app(self):
-                    try:
-                        result = await _run_async()
-                    finally:
-                        # Wait for the background tasks to be done. This needs to
-                        # go in the finally! If `_run_async` raises
-                        # `KeyboardInterrupt`, we still want to wait for the
-                        # background tasks.
-                        await self.cancel_and_wait_for_background_tasks()
+                # Make sure to set `_invalidated` to `False` to begin with,
+                # otherwise we're not going to paint anything. This can happen if
+                # this application had run before on a different event loop, and a
+                # paint was scheduled using `call_soon_threadsafe` with
+                # `max_postpone_time`.
+                self._invalidated = False
 
-                        # Set the `_is_running` flag to `False`. Normally this
-                        # happened already in the finally block in `run_async`
-                        # above, but in case of exceptions, that's not always the
-                        # case.
-                        self._is_running = False
+                loop = get_event_loop()
 
-                        # Also remove the Future again. (This brings the
-                        # application back to its initial state, where it also
-                        # doesn't have a Future.)
-                        self.future = None
+                if handle_sigint:
+                    loop.add_signal_handler(
+                        signal.SIGINT,
+                        lambda *_: loop.call_soon_threadsafe(
+                            self.key_processor.send_sigint
+                        ),
+                    )
 
-                    return result
-            finally:
                 if set_exception_handler:
-                    loop.set_exception_handler(previous_exc_handler)
+                    previous_exc_handler = loop.get_exception_handler()
+                    loop.set_exception_handler(self._handle_exception)
+
+                # Set slow_callback_duration.
+                original_slow_callback_duration = loop.slow_callback_duration
+                loop.slow_callback_duration = slow_callback_duration
+
+                try:
+                    with set_app(self), self._enable_breakpointhook():
+                        try:
+                            result = await _run_async()
+                        finally:
+                            # Wait for the background tasks to be done. This needs to
+                            # go in the finally! If `_run_async` raises
+                            # `KeyboardInterrupt`, we still want to wait for the
+                            # background tasks.
+                            await self.cancel_and_wait_for_background_tasks()
+
+                            # Also remove the Future again. (This brings the
+                            # application back to its initial state, where it also
+                            # doesn't have a Future.)
+                            self.future = None
+
+                        return result
+                finally:
+                    if set_exception_handler:
+                        loop.set_exception_handler(previous_exc_handler)
+
+                    if handle_sigint:
+                        loop.remove_signal_handler(signal.SIGINT)
+
+                    # Reset slow_callback_duration.
+                    loop.slow_callback_duration = original_slow_callback_duration
+
+            finally:
+                # Set the `_is_running` flag to `False`. Normally this happened
+                # already in the finally block in `run_async` above, but in
+                # case of exceptions, that's not always the case.
+                self._is_running = False
 
         return await _run_async2()
 
@@ -819,6 +860,7 @@ class Application(Generic[_AppResult]):
         self,
         pre_run: Optional[Callable[[], None]] = None,
         set_exception_handler: bool = True,
+        handle_sigint: bool = True,
         in_thread: bool = False,
     ) -> _AppResult:
         """
@@ -844,6 +886,8 @@ class Application(Generic[_AppResult]):
             `get_appp().create_background_task()`, so that unfinished tasks are
             properly cancelled before the event loop is closed. This is used
             for instance in ptpython.
+        :param handle_sigint: Handle SIGINT signal. Call the key binding for
+            `Keys.SIGINT`. (This only works in the main thread.)
         """
         if in_thread:
             result: _AppResult
@@ -853,7 +897,10 @@ class Application(Generic[_AppResult]):
                 nonlocal result, exception
                 try:
                     result = self.run(
-                        pre_run=pre_run, set_exception_handler=set_exception_handler
+                        pre_run=pre_run,
+                        set_exception_handler=set_exception_handler,
+                        # Signal handling only works in the main thread.
+                        handle_sigint=False,
                     )
                 except BaseException as e:
                     exception = e
@@ -918,6 +965,55 @@ class Application(Generic[_AppResult]):
 
         ensure_future(in_term())
 
+    @contextmanager
+    def _enable_breakpointhook(self) -> Generator[None, None, None]:
+        """
+        Install our custom breakpointhook for the duration of this context
+        manager. (We will only install the hook if no other custom hook was
+        set.)
+        """
+        if sys.version_info >= (3, 7) and sys.breakpointhook == sys.__breakpointhook__:
+            sys.breakpointhook = self._breakpointhook
+
+            try:
+                yield
+            finally:
+                sys.breakpointhook = sys.__breakpointhook__
+        else:
+            yield
+
+    def _breakpointhook(self, *a: object, **kw: object) -> None:
+        """
+        Breakpointhook which uses PDB, but ensures that the application is
+        hidden and input echoing is restored during each debugger dispatch.
+        """
+        app = self
+        # Inline import on purpose. We don't want to import pdb, if not needed.
+        import pdb
+        from types import FrameType
+
+        TraceDispatch = Callable[[FrameType, str, Any], Any]
+
+        class CustomPdb(pdb.Pdb):
+            def trace_dispatch(
+                self, frame: FrameType, event: str, arg: Any
+            ) -> TraceDispatch:
+                # Hide application.
+                app.renderer.erase()
+
+                # Detach input and dispatch to debugger.
+                with app.input.detach():
+                    with app.input.cooked_mode():
+                        return super().trace_dispatch(frame, event, arg)
+
+                # Note: we don't render the application again here, because
+                # there's a good chance that there's a breakpoint on the next
+                # line. This paint/erase cycle would move the PDB prompt back
+                # to the middle of the screen.
+
+        frame = sys._getframe().f_back
+        CustomPdb(stdout=sys.__stdout__).set_trace(frame)
+
     def create_background_task(
         self, coroutine: Awaitable[None]
     ) -> "asyncio.Task[None]":
@@ -1108,15 +1204,16 @@ class Application(Generic[_AppResult]):
         if _SIGTSTP is not None:
 
             def run() -> None:
+                signal = cast(int, _SIGTSTP)
                 # Send `SIGTSTP` to own process.
                 # This will cause it to suspend.
 
                 # Usually we want the whole process group to be suspended. This
                 # handles the case when input is piped from another process.
                 if suspend_group:
-                    os.kill(0, _SIGTSTP)
+                    os.kill(0, signal)
                 else:
-                    os.kill(os.getpid(), _SIGTSTP)
+                    os.kill(os.getpid(), signal)
 
             run_in_terminal(run)
 
@@ -1317,7 +1414,7 @@ def attach_winch_signal_handler(
 
     # Keep track of the previous handler.
     # (Only UnixSelectorEventloop has `_signal_handlers`.)
-    loop = asyncio.get_event_loop()
+    loop = get_event_loop()
     previous_winch_handler = getattr(loop, "_signal_handlers", {}).get(sigwinch)
 
     try:
