@@ -97,7 +97,7 @@ E = KeyPressEvent
 
 
 class Border:
-    " Box drawing characters. (Thin) "
+    "Box drawing characters. (Thin)"
     HORIZONTAL = "\u2500"
     VERTICAL = "\u2502"
     TOP_LEFT = "\u250c"
@@ -263,6 +263,10 @@ class TextArea:
 
         style = "class:text-area " + style
 
+        # If no height was given, guarantee height of at least 1.
+        if height is None:
+            height = D(min=1)
+
         self.window = Window(
             height=height,
             width=width,
@@ -317,11 +321,19 @@ class Label:
     """
     Widget that displays the given text. It is not editable or focusable.
 
-    :param text: The text to be displayed. (This can be multiline. This can be
-        formatted text as well.)
+    :param text: Text to display. Can be multiline. All value types accepted by
+        :class:`prompt_toolkit.layout.FormattedTextControl` are allowed,
+        including a callable.
     :param style: A style string.
     :param width: When given, use this width, rather than calculating it from
         the text size.
+    :param dont_extend_width: When `True`, don't take up more width than
+                              preferred, i.e. the length of the longest line of
+                              the text, or value of `width` parameter, if
+                              given. `True` by default
+    :param dont_extend_height: When `True`, don't take up more width than the
+                               preferred height, i.e. the number of lines of
+                               the text. `False` by default.
     """
 
     def __init__(
@@ -331,6 +343,7 @@ class Label:
         width: AnyDimension = None,
         dont_extend_height: bool = True,
         dont_extend_width: bool = False,
+        align: Union[WindowAlign, Callable[[], WindowAlign]] = WindowAlign.LEFT,
     ) -> None:
 
         self.text = text
@@ -352,9 +365,11 @@ class Label:
         self.window = Window(
             content=self.formatted_text_control,
             width=get_width,
+            height=D(min=1),
             style="class:label " + style,
             dont_extend_height=dont_extend_height,
             dont_extend_width=dont_extend_width,
+            align=align,
         )
 
     def __pt_container__(self) -> Container:
@@ -398,13 +413,19 @@ class Button:
             else:
                 return "class:button"
 
+        # Note: `dont_extend_width` is False, because we want to allow buttons
+        #       to take more space if the parent container provides more space.
+        #       Otherwise, we will also truncate the text.
+        #       Probably we need a better way here to adjust to width of the
+        #       button to the text.
+
         self.window = Window(
             self.control,
             align=WindowAlign.CENTER,
             height=1,
             width=width,
             style=get_style,
-            dont_extend_width=True,
+            dont_extend_width=False,
             dont_extend_height=True,
         )
 
@@ -429,7 +450,7 @@ class Button:
         ]
 
     def _get_key_bindings(self) -> KeyBindings:
-        " Key bindings for the Button. "
+        "Key bindings for the Button."
         kb = KeyBindings()
 
         @kb.add(" ")
@@ -524,7 +545,9 @@ class Frame:
                         fill(width=1, height=1, char=Border.BOTTOM_LEFT),
                         fill(char=Border.HORIZONTAL),
                         fill(width=1, height=1, char=Border.BOTTOM_RIGHT),
-                    ]
+                    ],
+                    # specifying height here will increase the rendering speed.
+                    height=1,
                 ),
             ],
             width=width,
